import 'dart:convert';
import 'dart:io';
import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';
import 'package:dio_logger/dio_logger.dart';
import 'package:flutter_mylibrary/config/const.dart';
import 'package:flutter_mylibrary/utils/utils.dart';
import 'exception/http_error.dart';
import 'http_data.dart';


/// 请求方法.
class Method {
  static final String get = "GET";
  static final String post = "POST";
  static final String put = "PUT";
  static final String head = "HEAD";
  static final String delete = "DELETE";
  static final String patch = "PATCH";
}


class EasyDio {

  //服务器路径

  static EasyDio _instance = EasyDio._internal();

  Dio _dio;

  factory EasyDio() => _instance;

  BaseOptions _baseOptions;

  Function _onError;

  Function _onFinish;

  Function _onStart;

  ValidateResponse _validateResponse;

  String _successCode;

  String _codeKey;

  String _msgKey;

  String _dataKey;

  Dio get dio => _dio;

  ///-----------------------------------------工具类初始化设置-------------------------------------

  EasyDio._internal();

  ///初始化公共属性
  ///
  /// [baseUrl] 地址前缀
  /// [connectTimeout] 连接超时赶时间
  /// [receiveTimeout] 接收超时赶时间
  /// [interceptors] 基础拦截器
  void init({String baseUrl,
    int connectTimeout,
    int receiveTimeout,
    String pem,
    String pKCSPath,
    String pKCSPwd,
    String codeKey = 'code',
    String msgKey = 'msg',
    String dataKey = 'data',
    String successCode = '0',
    ResponseType responseType,
    bool openLog = true,
    int retry = 3,
    Map<String, dynamic> headers,
    Map<String, dynamic> extra,
    String contentType,
    ValidateStatus validateStatus,
    bool followRedirects,
    Function onError,
    Function onFinish,
    Function onStart,
    ValidateResponse validateResponse,
    bool enableAutoCookie = false,
    ListFormat listFormat,
    List<Interceptor> interceptors
  }) {

    this._validateResponse = validateResponse;
    this._onStart = onStart;
    this._onError = onError;
    this._onFinish = onFinish;
    this._successCode = successCode;
    this._codeKey = codeKey;
    this._msgKey = msgKey;
    this._dataKey = dataKey;


    _baseOptions = BaseOptions(
      baseUrl: baseUrl,
      receiveTimeout: receiveTimeout,
      connectTimeout: connectTimeout,
      contentType: contentType,
      responseType: responseType??ResponseType.json,
      headers: headers??{},
      extra: extra??{},
      validateStatus: validateStatus,
      followRedirects: followRedirects,
      listFormat: listFormat
    );


    _dio = Dio(_baseOptions);


    if(openLog && isInDebugMode){
      _dio.interceptors.add(dioLoggerInterceptor);
    }


    if (retry != 0) {
      // dio.interceptors.add(element)
      // dio.interceptors.add(RetryInterceptor(
      //     dio: dio,
      //     options: RetryOptions(
      //       retries: retry, // Number of retries before a failure
      //       retryInterval: const Duration(seconds: 1), // Interval between each retry
      //     )
      // ));
    }



    ///-----------------------------------------证书相关信息-------------------------------------
    (_dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate = (HttpClient client) {
      client.badCertificateCallback = (X509Certificate cert, String host, int port) {
        if (isNotEmpty(pem)) {
          if (cert.pem == pem) {
            // 证书一致，则放行
            return true;
          }else{
            return false;
          }
        }else{
          return true;
        }
      };
    };

    if (isNotEmpty(pKCSPath) && isNotEmpty(pKCSPwd)) {
      (_dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate = (HttpClient client) {
        SecurityContext sc = new SecurityContext();
        //file为证书路径
        sc.setTrustedCertificates(pKCSPath, password: pKCSPwd);
        HttpClient httpClient = new HttpClient(context: sc);
        return httpClient;
      };
    }


  }




  ///-----------------------------------------请求方法封装-------------------------------------
  Future<HttpData> request(
      String method,
      String path, {
        Options options,
        String cacheKey,
        bool isFormData = false,
        bool publicMethod = true,
        data,
        String contentType,
        CancelToken cancelToken,
        Function onStart,
        Function onError,
        Function onFinish,
        ValidateResponse validateResponse,
        String successCode,
        String codeKey,
        String msgKey,
        String dataKey,
        ProgressCallback onReceiveProgress,
        ProgressCallback onSendProgress
      }) async{


    HttpData _httpData = HttpData();
    String _code;
    String _msg;
    var _resultData;

    try{
      if (onStart != null){
        onStart();
      }else if(this._onStart != null){
        if (publicMethod) {
          this._onStart();
        }
      }


      Response response;


      if (isFormData) {
        data = FormData.fromMap(data);
      }


      if(isNotEmpty(options) && isEmpty(options.method)){
        options.method = method;
      }


      if (method == Method.get) {
        response = await _dio.request(
            path,
            queryParameters: data,
            options: options,
            cancelToken: cancelToken,
            onReceiveProgress: (int sent, int total){
              onReceiveProgress?.call(sent, total);
            },
            onSendProgress: (int sent, int total){
              onSendProgress?.call(sent, total);
            }
        );
      }else{
        response = await _dio.request(
            path,
            data: data,
            options: options,
            cancelToken: cancelToken,
            onReceiveProgress: (int sent, int total){
              onReceiveProgress?.call(sent, total);
            },
            onSendProgress: (int sent, int total){
              onSendProgress?.call(sent, total);
            }
        );
      }



      String mCodeKey = codeKey??_codeKey;
      String mMsgKey = msgKey??_msgKey;
      String mDataKey = dataKey??_dataKey;
      String mSuccessCode = successCode??_successCode;


      if (response.data is List) {
        _resultData = response.data;
        _code = successCode??_successCode;
      }else{
        Map<String, dynamic> resJson;

        try{
          if(response.data is Map){
            resJson = response.data;
          }else{
            resJson = json.decode(response.data);
          }
        }catch(e){
          throw HttpError(HttpErrorType.DataError, PARSE_ERROR_MSG, cause: e, logMsg: '数据解析错误');
        }

        if(resJson.containsKey(mCodeKey)){
          _code = resJson[mCodeKey].toString();
        }

        if(resJson.containsKey(mMsgKey)){
          _msg = resJson[mMsgKey].toString();
        }



        if (isNotEmpty(validateResponse)) {
          validateResponse(response.data);
        }else if(isNotEmpty(_validateResponse)){
          _validateResponse(response.data);
        }



        if(_code != mSuccessCode){
          throw HttpError(HttpErrorType.ResultError, isNotEmpty(_msg)? _msg : HTTP_INCORRECT_MSG, logMsg: '请求code错误，成功code: $mSuccessCode, 请求code: $_code');
        }else{
          if(resJson.containsKey(mDataKey)){
            _resultData = resJson[mDataKey];
            if(isEmpty(_resultData)){
              throw HttpError.emptyDataError(_resultData);
            }
          }else{
            throw HttpError.emptyDataError(_resultData);
          }
        }
      }
    }catch(e){
      HttpError error;
      if (e is DioError) {
        error = HttpError.dioError(e);
      }else if(e is HttpError){
        error = e;
      }else{
        error = HttpError.unknown(e);
      }

      if(onError != null){
        onError.call(error);
      }else if (this._onError != null) {
        if (publicMethod) {
          this._onError.call(error);
        }
      }
      _httpData.httpError = error;
    }finally{
      if (isNotEmpty(onFinish)) {
        onFinish();
      } else if(isNotEmpty(this._onFinish) && publicMethod){
        this._onFinish();
      }
    }

    _httpData.code = _code;
    _httpData.msg = _msg;
    _httpData.data = _resultData;

    return _httpData;
  }



  ///-----------------------------------------上传文件方法封装-------------------------------------------------------------
  void fileUpload({
    MultipartFile multipartFile,
    String url,
    String filePath,
    String key = 'file',
    String fileName = '',
    FormData formData,
    ProgressCallback onSendProgress,
    ProgressCallback onReceiveProgress,
    Function onStart,
    Function onError,
    Function onFinish,
    bool publicMethod = true}) async{

    MultipartFile file = multipartFile??await MultipartFile.fromFile(filePath, filename: fileName);


    try{
      await _dio.post(
        url,
        data: formData??FormData.fromMap({
          key: file
        }),
        onSendProgress: (int sent, int total){
          onSendProgress?.call(sent, total);
        },
        onReceiveProgress: (int sent, int total){
          onReceiveProgress?.call(sent, total);
        },
      );
    }on DioError catch(e){
      throw HttpError.dioError(e);
    }catch(e){
      if (e is HttpError) {
        if(onError != null){
          onError.call(e);
        }else if (publicMethod) {
          this._onError?.call(e);
        }

        if (isNotEmpty(onFinish)) {
          onFinish();
        } else if(publicMethod){
          this._onFinish?.call();
        }
      }
    }
  }


  void filesUpload({
    List<MultipartFile> multipartFile,
    String url,
    String filePath,
    String key = 'file',
    String fileName = '',
    FormData formData,
    ProgressCallback onSendProgress,
    ProgressCallback onReceiveProgress,
    Function onStart,
    Function onError,
    Function onFinish,
    bool publicMethod = true}) async{


    try{
      await _dio.post(
        url,
        data: formData??FormData.fromMap({
          key: multipartFile
        }),
        onSendProgress: (int sent, int total){
          onSendProgress?.call(sent, total);
        },
        onReceiveProgress: (int sent, int total){
          onReceiveProgress?.call(sent, total);
        },
      );
    }on DioError catch(e){
      throw HttpError.dioError(e);
    }catch(e){
      if (e is HttpError) {
        if(onError != null){
          onError.call(e);
        }else if (publicMethod) {
          this._onError?.call(e);
        }

        if (isNotEmpty(onFinish)) {
          onFinish();
        } else if(publicMethod){
          this._onFinish?.call();
        }
      }
    }
  }


  ///-----------------------------------------下载文件方法封装-------------------------------------------------------------
  void download({
    String urlPath = '',
    String savePath = '',
    bool deleteOnError = true,
    String lengthHeader = Headers.contentLengthHeader,
    ProgressCallback onReceiveProgress,
    data,
    Map<String, dynamic> queryParameters,
    CancelToken cancelToken,
    Options options,
    Function onStart,
    Function onError,
    Function onFinish,
    bool publicMethod = true}) async{

    try{
      await _dio.download(
        urlPath,
        savePath,
        deleteOnError: deleteOnError,
        lengthHeader: lengthHeader,
        data: data,
        options: options,
        queryParameters: queryParameters,
        cancelToken: cancelToken,
        onReceiveProgress: (int sent, int total){
          onReceiveProgress?.call(sent, total);
        },
      );
    }on DioError catch(e){
      throw HttpError.dioError(e);
    }catch(e){
      if (e is HttpError) {
        if(onError != null){
          onError.call(e);
        }else if (publicMethod) {
          this._onError?.call(e);
        }

        if (isNotEmpty(onFinish)) {
          onFinish();
        } else if(publicMethod){
          this._onFinish?.call();
        }
      }
    }
  }




  ///-----------------------------------------dio部分设置-------------------------------------


  void setCookie(String cookie){
    Map<String, dynamic> _headers = new Map();
    _headers["Cookie"] = cookie;
    _dio.options.headers.addAll(_headers);

  }


  void setHeaders(Map<String, dynamic> headers){
    _dio.options.headers.addAll(headers);
  }

  void clearHeaders(Map<String, dynamic> headers){
    _dio.options.headers.clear();
  }

  ///修改默认url
  void changeBaseUrl(String baseUrl){
    _dio.options.baseUrl = baseUrl;
  }

}





